use std::{fs, path::PathBuf, process::exit, str::FromStr};
mod rsa;
use log::{*};
use fern::colors::Color;
use fern::colors::ColoredLevelConfig;
use log;
use clap::{self, Parser, Subcommand, ValueEnum};
use log::info;
use num_bigint_dig::BigUint;

#[derive(Parser)]
#[command(about = r#"
 ____  ____    _    
|  _ \/ ___|  / \   
| |_) \___ \ / _ \  
|  _ < ___) / ___ \ 
|_| \_\____/_/   \_\

A nice RSA implementation thats definetly not vulnerable to a lot of crypto magic"#)]
pub struct Args{
    #[command(subcommand)]
    mode: Mode,
    #[arg(short, long, help = "How many log messages do you want to see?")]
    log_level: Option<log::LevelFilter>,
}

#[derive(Clone, Subcommand)]
pub enum Mode{
    #[command(about = "generate a new keypair")]
    Generate{
        #[arg(short, long, value_name = "Key Size", help = "The size of the RSA key in bits. Must be a power of 2")]
        bitsize: Option<usize>,
    },
    #[command(about = "Encrypt a string using a public key")]
    Encrypt{
        #[arg(value_name = "Public Key", help = "The path to the public key file")]
        pub_key: PathBuf,
        #[arg(value_name = "Message")]
        message: String,
    },
    #[command(about = "Decrypt a string using a private key")]
    Decrypt{
        #[arg[value_name = "Private Key", help = "The path to the private key file"]]
        priv_key: PathBuf,
        #[arg(value_name = "Message File", help = "The path to the message file")]
        message: PathBuf,
    },
}

/// Key file format
/// <value of n>:<rest of key (public or private)>
fn parse_key_from_string(key: String) -> (BigUint, BigUint){
    info!("Splitting input message into blocks");
    let parts = key.split(":").map(|s| s.to_string()).collect::<Vec<String>>();
    info!("{} Blocks", parts.len());

    if parts.len() != 2{
        error!("Failed to parse key file. More than 2 segments provided");
        exit(-1);
    };

    let n = BigUint::from_str(&parts[0]);
    let k = BigUint::from_str(&parts[1]);

    if k.is_err() || n.is_err(){
        error!("Failed to parse keys. Format doesnt appear to be in integer format");
        exit(-1);
    };

    info!("Keyfile parsed successfully");

    (n.unwrap(), k.unwrap())
}

fn main() {
    let args = Args::parse();
    match setup_logger(&args){
        Ok(_) => (),
        Err(e) => println!("Faiiled to setup logger {e}. Proceeding without"),
    };

    match args.mode{
        Mode::Generate {bitsize} => generate_cli(bitsize),
        Mode::Decrypt { priv_key, message } => decrypt_cli(priv_key, message),
        Mode::Encrypt { pub_key, message } => encrypt_cli(pub_key, message),
    }
}

fn generate_cli(bitsize: Option<usize>){
    info!("Generating keypair");
    let (n, e, d) = rsa::generate_magics(bitsize);

    let public_keypair = format!("{n}:{e}");
    let private_keypair = format!("{n}:{d}");
    let pub_res = fs::write("public_key", public_keypair);
    let priv_res = fs::write("private_key", private_keypair);

    if pub_res.is_err() || priv_res.is_err(){
        error!("Failed to write keys to file");
        exit(-1);
    }
    info!("Saved private key to \"private_key\"");
    info!("Saved public key to \"public_key\"");
}

fn encrypt_cli(pub_key_file: PathBuf, message: String){
    let key_file = match fs::read_to_string(pub_key_file){
        Ok(s) => s,
        Err(e) => {error!("Error: {e}"); exit(-1);}
    };

    let (n, e) = parse_key_from_string(key_file);

    if fs::write("message.enc", rsa::encrypt(message, &e, &n)).is_err(){
        error!("Failed to write message to file");
        exit(-1);
    };
}

fn decrypt_cli(priv_key_file: PathBuf, message: PathBuf){
    let key_file = match fs::read_to_string(priv_key_file){
        Ok(s) => s,
        Err(e) => {error!("Error: {e}"); exit(-1);}
    };
    let message_content = match fs::read_to_string(message){
        Ok(s) => s,
        Err(e) => {error!("Error: {e}"); exit(-1);}
    };

    let (n, d) = parse_key_from_string(key_file);
    println!("{}", rsa::decrypt(message_content, &d, &n));
}

fn setup_logger(args: &Args) -> anyhow::Result<()>{
    let colors = ColoredLevelConfig::new()
    .trace(Color::Blue).info(Color::Green).warn(Color::Magenta).error(Color::Red);
    fern::Dispatch::new()
    // Perform allocation-free log formatting
    .format(move |out, message, record| {
        out.finish(format_args!(
            "[{} {} {}] {}",
            humantime::format_rfc3339(std::time::SystemTime::now()),
            colors.color(record.level()),
            record.target(),
            message
        ))
    })
    // Add blanket level filter -
    .level(args.log_level.unwrap_or(log::LevelFilter::Warn))
    // - and per-module overrides
    .level_for("hyper", log::LevelFilter::Info)
    // Output to stdout, files, and other Dispatch configurations
    .chain(std::io::stdout())
    // Apply globally
    .apply()?;

    Ok(())
}