use std::{borrow::Cow, char::DecodeUtf16, fs::read_link, str::FromStr};
use num_bigint_dig::{self, BigUint, IntoBigUint, UniformBigUint};
use rand::{self, distributions::uniform::UniformSampler, thread_rng};
use log::*;

fn gcd(mut b1: BigUint, mut b2: BigUint) -> BigUint {
    let mut tmp: BigUint;
    loop {
        // see if 5 is cleanly divisible by 24 (no)
        // tmp = 5
        tmp = b1 % b2.clone();
        if tmp == 0_u8.into() {
            break;
        }
        // b1 = 4
        b1 = b2;
        // b2 = 1
        b2 = tmp;
    }
    return b2;
}

fn generate_bitarray(length: usize) -> Vec<u8>{
    let mut out = vec![];
    for _ in 0..length{
        out.push(rand::random::<u8>())
    }
    out
}

/// Returns (n, e, d)
pub fn generate_magics(bitsize: Option<usize>) -> (BigUint, BigUint, BigUint){
    let bitsize = bitsize.unwrap_or(2048);
    assert!(bitsize.is_power_of_two(), "Bizsize is not a power of 2 (this is your fault)");
    let p;
    let q;

    log::warn!("These implementations of random number generation are NOT cryptographically safe.");
    let bitarray = generate_bitarray(bitsize/8);
    let start = BigUint::from_bytes_le(&bitarray);
    p = num_bigint_dig::prime::next_prime(&start);
    let bitarray = generate_bitarray(bitsize/8);
    let start = BigUint::from_bytes_le(&bitarray);
    q = num_bigint_dig::prime::next_prime(&start);

    let n = p.clone()*q.clone();
    let f = (p.clone()-1_u8)*(q.clone()-1_u8);
    
    let e = {
        let haha = UniformBigUint::new::<BigUint, BigUint>(2_u8.into(), f.clone());
        let mut rng = thread_rng();
        let mut sample = haha.sample(&mut rng);
        while sample < f && gcd(f.clone(), sample.clone()) != 1_u8.into(){
            sample = haha.sample(&mut rng);
        }

        sample
    };

    //let lcm = f.clone()/gcd(p.clone()-1_u8, q.clone()-1_u8);
    let d = num_bigint_dig::algorithms::mod_inverse(Cow::Owned(e.clone()), Cow::Owned(f.clone())).unwrap().into_biguint().unwrap();
    info!("Generated keypair of {} bits", bitsize);
    (n, e, d)
}

pub fn encrypt(message: String, e: &BigUint, n: &BigUint) -> String{
    let mut blocks = vec![];
    let message = message.as_bytes();
    let max_len = n.to_bytes_le().len()-1;
    for i in 0..(&message.len()/max_len){
        blocks.push(&message[i*max_len..(i+1)*max_len]);
    }
    if message.len() % max_len != 0 {
        blocks.push(&message[((message.len() / max_len) * max_len)..])
    }

    let mut out = String::new();
    for block in blocks{
        let bytes = block;
        let mut tmp_out = BigUint::from_bytes_le(bytes);
        tmp_out = tmp_out.modpow(&e, &n);
        out += &format!("{}-", tmp_out);
    }

    
    out[..out.len()-1].to_string()
}

pub fn decrypt(message: String, d: &BigUint, n: &BigUint) -> String{
    let blocks = message.split("-").into_iter();
    let mut bytes = vec![];
    for block in blocks{
        let numba = BigUint::from_str(block).unwrap();
        bytes.extend_from_slice(&numba.modpow(&d, &n).to_bytes_le());
    }
    String::from_utf8_lossy(&bytes).into_owned()
}

pub fn stupid_decrypt(mut input: String, d: BigUint, n: BigUint) -> String{
    let mut out = vec![];
    for i in input.split("-"){
        if i.len() == 0{
            continue;
        }
        let encrypted_bruh = BigUint::from_str(i).unwrap();
        let decrypted = encrypted_bruh.modpow(&d, &n);

        let decrypted = decrypted.to_string();
        let mut blocks = vec![];
        for i in 0..(&decrypted.len()/3){
            blocks.push(u8::from_str(&decrypted[i*3..(i+1)*3]).unwrap());
        }
        out.push(String::from_utf8_lossy(&blocks[..]).into_owned());
    }

    out.join("")
}

pub fn stupid_encrypt(mut message: String, e: BigUint, n: BigUint) -> String{
    let mut message = message.to_lowercase();
    message += &" ".repeat(4 - (message.len() % 4));

    let mut blocks = vec![];
    for i in 0..(&message.len()/4){
        blocks.push(&message[i*4..(i+1)*4]);
    }

    let mut hahas = vec![];

    for block in blocks{
        let mut haha = String::new();
        for b in block.bytes(){
            haha += &format!("{:0>3}", u8::from(b));
        }
        hahas.push(haha);
    }

    let mut out = String::new();
    for haha in hahas{
        let message = BigUint::from_str(&haha).unwrap();
        let encrypted = message.modpow(&e, &n);
        out += "-";
        out += &encrypted.to_string();
    }

    out
}